//
//  AmpDatabaseManager.swift
//  AmpDownloader
//
//  Created by David Cortes on 1/25/19.
//  Copyright © 2019 Miguel Hernandez. All rights reserved.
//

import Foundation
import AVFoundation

internal class AmpAssetDatabaseManager:NSObject {
    
    static var sharedManager:AmpAssetDatabaseManager {
        get {
            if instance == nil {
                instance = AmpAssetDatabaseManager()
            }
            
            return instance!
        }
    }
    
    fileprivate static var instance:AmpAssetDatabaseManager?
    
    fileprivate var assetDownloadURLSession: AVAssetDownloadURLSession!
    
    fileprivate var restoreSession:URLSession!
    
    override private init() {
        super.init()
        let backgroundConfiguration = URLSessionConfiguration.background(withIdentifier: "AmpDownloader")
        assetDownloadURLSession = AVAssetDownloadURLSession(configuration: backgroundConfiguration, assetDownloadDelegate: self, delegateQueue: OperationQueue.main)
    }
    
    func restoreManager() {
        
        print("restoring")
        
        assetDownloadURLSession.getAllTasks { [weak self] tasks in

            print("tasks #: \(tasks.count)")
            for task in tasks {
                guard let assetDownloadTask = task as? AVAggregateAssetDownloadTask else {
                    break
                }

                let reason = task.error?._userInfo?[NSURLErrorBackgroundTaskCancelledReasonKey] as? Int
                let code = task.error?._code

                if reason == NSURLErrorCancelledReasonUserForceQuitApplication && code == NSURLErrorCancelled {
                    print("Something happened, restarting download.")
                    guard let taskId = assetDownloadTask.taskDescription else {
                        return
                    }
                    
                    if let download = AmpWritter.getDownload(id:taskId, withCanceledDownloads: true) {
                        self?.restartDownload(download:download)
                    }
                } else {
                    if let download = AmpWritter.getDownload(id:assetDownloadTask.taskDescription ?? "") {
                        print("Download retrieved")
                        if download._isPaused {
                            assetDownloadTask.suspend()
                        } else {
                            assetDownloadTask.resume()
                        }
                    } else {
                        print("Download not retrieved")
                    }
                }
            }
        }
        
        AmpWritter.removeCanceledDownloads()
    }
    
    func restartDownload(download:AmpDownload) {
        if !AmpWritter.deleteDownload(download: download) {
            print("Couldn't delete")
        }
        
        if let _ = downloadAsset(url: download._url, title: download.title, assetArtworkData: download._assetArtworkData, pause: download.isPaused) {
            print("Download restored")
        } else {
            print("Couldn't restore download")
        }
        
    }
    
    
    func downloadAsset(url:String, title:String, assetArtworkData:Data? = nil, pause:Bool = false) -> AmpDownload? {
        
        let download = AmpDownload(title: title, url: url, assetArtworkData: assetArtworkData)
        let preferredMediaSelection = download.asset!.preferredMediaSelection
        
        guard let task = assetDownloadURLSession.aggregateAssetDownloadTask(with: download.asset!, mediaSelections: [preferredMediaSelection], assetTitle: title, assetArtworkData: assetArtworkData, options: nil) else { return nil }
        
        task.taskDescription = download._id
        
        download._isPaused = pause
        if pause {
            task.suspend()
        } else {
            task.resume()
        }
        
        AmpDownloadManager.notifyDownloadStart(asset: download)
        
        let _ = AmpWritter.insertDownload(download:download)
        
        return download
        
    }
    
    /// Deletes an Asset on disk if possible.
    func deleteAsset(_ asset:AmpDownload) {
        let userDefaults = UserDefaults.standard
        
        do {
            if let localFileLocation = localAssetForStream(withName: asset.title)?.asset?.url {
                try FileManager.default.removeItem(at: localFileLocation)
                
                userDefaults.removeObject(forKey: asset.title)
            }
        } catch {
            print("An error occured deleting the file: \(error)")
        }
    }
    
    /// Resumes an AVAssetDownloadTask given an Asset.
    func resumeDownload(for asset: AmpDownload) {
        assetDownloadURLSession.getAllTasks {tasks in
            if let task = tasks.filter({ $0.taskDescription == asset._id }).first {
                task.resume()
            }
        }
    }
    
    /// Pause an AVAssetDownloadTask given an Asset.
    func pauseDownload(for asset: AmpDownload) {
        assetDownloadURLSession.getAllTasks { tasks in
            if let task = tasks.filter({ $0.taskDescription == asset._id }).first {
                task.suspend()
            }
        }
    }
    
    /// Cancels an AVAssetDownloadTask given an Asset.
    func cancelDownload(for asset: AmpDownload) {
        assetDownloadURLSession.getAllTasks { tasks in
            if let task = tasks.filter({ $0.taskDescription == asset.id }).first {
                task.cancel()
            }
        }
    }
    
    /// Return the display names for the media selection options that are currently selected in the specified group
    func displayNamesForSelectedMediaOptions(_ mediaSelection: AVMediaSelection) -> String {
        
        var displayNames = ""
        
        guard let asset = mediaSelection.asset else {
            return displayNames
        }
        
        // Iterate over every media characteristic in the asset in which a media selection option is available.
        for mediaCharacteristic in asset.availableMediaCharacteristicsWithMediaSelectionOptions {
            /*
             Obtain the AVMediaSelectionGroup object that contains one or more options with the
             specified media characteristic, then get the media selection option that's currently
             selected in the specified group.
             */
            guard let mediaSelectionGroup =
                asset.mediaSelectionGroup(forMediaCharacteristic: mediaCharacteristic),
                let option = mediaSelection.selectedMediaOption(in: mediaSelectionGroup) else { continue }
            
            // Obtain the display string for the media selection option.
            if displayNames.isEmpty {
                displayNames += " " + option.displayName
            } else {
                displayNames += ", " + option.displayName
            }
        }
        
        return displayNames
    }
    
    /// Returns an Asset pointing to a file on disk if it exists.
    func localAssetForStream(withName name: String) -> AmpDownload? {
        let userDefaults = UserDefaults.standard
        guard let localFileLocation = userDefaults.value(forKey: name) as? Data else { return nil }
        
        var asset: AmpDownload?
        var bookmarkDataIsStale = false
        do {
            let url = try URL(resolvingBookmarkData: localFileLocation,
                              bookmarkDataIsStale: &bookmarkDataIsStale)
            
            if bookmarkDataIsStale {
                fatalError("Bookmark data is stale!")
            }
            
            asset = AmpDownload(title: name, url: url!.absoluteString)
            
            return asset
        } catch {
            fatalError("Failed to create URL from bookmark with error: \(error)")
        }
    }
    
}

@available(iOS 11.0, *)
extension AmpAssetDatabaseManager:AVAssetDownloadDelegate {
    func urlSession(_ session: URLSession, aggregateAssetDownloadTask: AVAggregateAssetDownloadTask, willDownloadTo location: URL) {
        guard let taskId = aggregateAssetDownloadTask.taskDescription else {
            return
        }
        
        let _ = AmpWritter.setDownloadLocation(id: taskId, location:location)
    }
    
    func urlSession(_ session: URLSession, aggregateAssetDownloadTask: AVAggregateAssetDownloadTask, didCompleteFor mediaSelection: AVMediaSelection) {
        /*
         This delegate callback provides an AVMediaSelection object which is now fully available for
         offline use. You can perform any additional processing with the object here.
         */
        guard let taskId = aggregateAssetDownloadTask.taskDescription, let downloadedAsset = AmpWritter.getDownload(id:taskId) else {
            return
        }
        
        aggregateAssetDownloadTask.resume()
        
        AmpDownloadManager.notifyStateChanged(asset: downloadedAsset, mediaOption: displayNamesForSelectedMediaOptions(mediaSelection))
    }
    
    func urlSession(_ session: URLSession, aggregateAssetDownloadTask: AVAggregateAssetDownloadTask, didLoad timeRange: CMTimeRange, totalTimeRangesLoaded loadedTimeRanges: [NSValue], timeRangeExpectedToLoad: CMTimeRange, for mediaSelection: AVMediaSelection) {
        guard let taskId = aggregateAssetDownloadTask.taskDescription else {
            print("MIGUEL no task id found")
            return
        }
        
        guard let downloadedAsset = AmpWritter.getDownload(id:taskId) else {
            print("MIGUEL no asset found")
            return
        }
        
        if downloadedAsset.isPaused && aggregateAssetDownloadTask.state == .running {
            let _ = AmpWritter.resumeDownload(download:downloadedAsset)
            AmpDownloadManager.notifyDownloadResume(asset: downloadedAsset)
        }
        
        
        var percentComplete = 0.0
        for value in loadedTimeRanges {
            let loadedTimeRange: CMTimeRange = value.timeRangeValue
            percentComplete +=
                loadedTimeRange.duration.seconds / timeRangeExpectedToLoad.duration.seconds
        }
        AmpDownloadManager.notifyDownloadProgressChange(asset: downloadedAsset, progress: Float(percentComplete))
        
        if aggregateAssetDownloadTask.state == .suspended && !downloadedAsset.isPaused {
            let _ = AmpWritter.pauseDownload(download:downloadedAsset)
            AmpDownloadManager.notifyDownloadPause(asset: downloadedAsset)
        }
    }
    
    func urlSession(_ session: URLSession, task: URLSessionTask, didCompleteWithError error: Error?) {
        guard let taskId = task.taskDescription, let asset = AmpWritter.getDownload(id:taskId) else {
            return
        }
        
        if let error = error as NSError? {
            switch (error.domain, error.code) {
            case (NSURLErrorDomain, NSURLErrorCancelled):
                let _ = AmpWritter.deleteDownload(download:asset)
                AmpDownloadManager.notifyDownloadCancelled(asset: asset)
            case (NSURLErrorDomain, NSURLErrorUnknown):
                let _ = AmpWritter.failDownload(download:asset)
                AmpDownloadManager.notifyDownloadError(asset: asset, error: error, message:"Downloading HLS streams is not supported in the simulator.")
            default:
                let _ = AmpWritter.failDownload(download:asset)
                AmpDownloadManager.notifyDownloadError(asset: asset, error: error, message:"An unexpected error occured \(error.domain)")
            }
        } else {
            let _ = AmpWritter.completeDownload(download:asset)
            AmpDownloadManager.notifyDownloadCompleted(asset: asset)
        }
    }
}
